import base64
import binascii

from datetime import date, datetime
from typing import Dict
from typing import List

import dacite
from requests.models import Response

from supplier_api_tester.v2.exceptions import FailedTest
from supplier_api_tester.v2.models import ApiError, Booking, DailyVariants, Product, Reservation


def check_base64(item):
    try:
        return base64.b64encode(base64.b64decode(item)).decode() == item
    except binascii.Error:
        return False


def format_error_message(e: Exception) -> str:
    '''
    Method for enhancing the error messages from Dacite.
    '''
    if isinstance(e, dacite.exceptions.UnexpectedDataError):
        return str(e).replace(
            'to any data class field',
            'to any field from the specification'
        )
    return str(e)


def booking_pdf_validator(booking: Booking, raw_response: Response):
    if booking.barcode_scope == 'order':
        if not check_base64(booking.barcode):
            raise FailedTest(
                message="Error while decoding (base64) PDF voucher for the order",
                response=raw_response
        )
    elif booking.barcode_scope == 'ticket':
        for variant, barcodes in booking.tickets.items():
            for barcode in barcodes:
                if not check_base64(barcode):
                    raise FailedTest(
                        message="Error while decoding (base64) PDF voucher for the ticket",
                        response=raw_response
                    )


def parse_availability_variants(raw_response: Response, response: Dict) -> List[DailyVariants]:
    """Getting and testing response from the /availability endpoint"""
    if type(response) is not dict:
        raise FailedTest(
            message='The response should be a JSON Object',
            response=raw_response,
        )

    available_variants: List[DailyVariants] = []
    try:
        for day, day_availability in response.items():
            # extract timeslot component from the day value: YYYY-MM-DDTHH:MM
            date_time_value = datetime.fromisoformat(day)
            available_variants.append(
                dacite.from_dict(
                    data_class=DailyVariants,
                    data={
                        'date': str(date_time_value.date()),
                        'timeslot': f'{date_time_value.time().hour}:{date_time_value.time().minute}',
                        'available_tickets': day_availability.get('available_tickets', 0),
                        'variants': [
                            {**v}
                            for v in day_availability.get('variants', [])
                        ]
                    },
                    config=dacite.Config(
                        type_hooks={date: date.fromisoformat},
                        strict=True,
                    )
                )
            )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /availability endpoint: {format_error_message(e)}',
            response=raw_response,
        )
    return available_variants


def get_products(raw_response: Response, response: Dict) -> List[Product]:
    """Getting and testing response from the /products endpoint"""
    if type(response) is not list:
        raise FailedTest(
            message='The response should be a JSON Array',
            response=raw_response,
        )
    try:
        products = [
            dacite.from_dict(
                data_class=Product,
                data=product,
            )
            for product in response
        ]
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /products endpoint: {format_error_message(e)}',
            response=raw_response,
        )
    return products


def get_reservation(raw_response: Response, response) -> Reservation:
    """Getting and testing response from the /reservation endpoint"""
    if type(response) is not dict:
        raise FailedTest(
            message='The response should be a JSON Object',
            response=raw_response,
        )
    try:
        return dacite.from_dict(
            data_class=Reservation,
            data=response,
            config=dacite.Config(
                type_hooks={datetime: datetime.fromisoformat},
                strict=True,
            )
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /reservation endpoint: {format_error_message(e)}',
            response=raw_response,
        )


def get_booking(raw_response: Response, response) -> Booking:
    """Getting and testing response from the /booking endpoint"""
    if type(response) is not dict:
        raise FailedTest(
            message='The response should be a JSON Object',
            response=raw_response,
        )
    try:
        booking = dacite.from_dict(
            data_class=Booking,
            data=response,
            config=dacite.Config(strict=True)
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /booking endpoint: {format_error_message(e)}',
            response=raw_response,
        )

    if booking.barcode_format not in ('QRCODE', 'CODE128', 'CODE39', 'ITF', 'DATAMATRIX', 'EAN13', 'PDF'):

        raise FailedTest(
            message=f'Incorrect barcode format ({booking.barcode_format})',
            response=raw_response,
        )

    if booking.barcode_format == 'PDF':
        booking_pdf_validator(booking, raw_response)

    if booking.barcode_scope not in ('order', 'ticket'):
        raise FailedTest(
            message=f'Incorrect value in the barcode_scope field: {booking.barcode_scope}',
            response=raw_response,
        )
    if booking.barcode_scope == 'order' and not booking.barcode:
        raise FailedTest(
            message='Barcode for the whole order is empty',
            response=raw_response,
        )
    if booking.barcode_scope == 'ticket' and not booking.tickets:
        raise FailedTest(
            message='Tickets Array is empty',
            response=raw_response,
        )
    return booking


def get_api_error(raw_response: Response, response) -> ApiError:
    """Unpacking 400 error JSON structure"""
    if raw_response.ok:
        raise FailedTest(
            message=f'Expected HTTP 400 but got HTTP {response.get("status_code")} instead.',
            response=raw_response,
        )
    if type(response) is not dict:
        raise FailedTest(
            message='400 error response should be a JSON Object',
            response=raw_response,
        )
    try:
        return dacite.from_dict(
            data_class=ApiError,
            data=response,
            config=dacite.Config(strict=True),
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect response format for 400 error: {format_error_message(e)}',
            response=raw_response,
        )



# If multiple checks can be done using the same response
# then they should be done under a single test case.

from datetime import datetime, timedelta

from supplier_api_tester.v2.decorators import test_wrapper
from supplier_api_tester.v2.exceptions import FailedTest
from supplier_api_tester.v2.models import TestResult
from ..utils import availability


@test_wrapper
def test_next_30_days(api_url, api_key, product_id, version=2):
    """Checking availability for the next 30 days"""
    start_date = datetime.utcnow().date()
    end_date = start_date + timedelta(days=30)
    days, raw_response = availability.get_availability(
        api_url,
        api_key,
        product_id,
        version,
        start_date,
        end_date,
    )

    # non-zero availability check
    max_tickets_sum = sum(day.available_tickets for day in days)
    if max_tickets_sum <= 0:
        raise FailedTest(
            message='There is no availability for next 30 days',
            response=raw_response,
        )

    # variant IDs uniqueness
    variant_ids = set()
    variant_name_id_map = {}
    variant_counter = 0
    days_counter = 0
    for day in days:
        days_counter += 1
        for variant in day.variants:
            # checking if one variant name always have the same ID
            if variant.name in variant_name_id_map:
                if variant.id != variant_name_id_map[variant.name]:
                    raise FailedTest(
                        message=f'Variant {variant.name} should always have the same ID.',
                        response=raw_response,
                    )
            else:
                variant_name_id_map[variant.name] = variant.id

            variant_counter += 1
            variant_ids.add(variant.id)

        # checking number of unique variants in timespan of 7 days
        if days_counter == 7:
            days_counter = 0
            if len(variant_ids) > 20:
                return TestResult(
                    status=1,
                    message=(
                        'More then 20 unique variants were found in a timespan of 7 days. '
                        'Make sure that this is not an error. Variants should not be unique for '
                        'each day.'
                    ),
                )
            variant_ids = set()

    return TestResult()


@test_wrapper
def test_missing_api_key(api_url, api_key, product_id, version=2):
    """Request without API-Key"""
    return availability.test_missing_api_key(api_url, api_key, product_id, 'availability', version=version)


@test_wrapper
def test_incorrect_api_key(api_url, api_key, product_id, version=2):
    """Request with incorrect API-Key"""
    return availability.test_incorrect_api_key(api_url, api_key, product_id, 'availability', version=version)


@test_wrapper
def test_missing_argument_error(api_url, api_key, product_id, version=2):
    """Testing missing argument errors"""
    return availability.test_missing_argument_error(api_url, api_key, product_id, 'availability', version=version)


@test_wrapper
def test_error_for_non_existing_product(api_url, api_key, product_id, version=2):
    """Testing availability for non existing product"""
    return availability.test_error_for_non_existing_product(
        api_url, api_key, product_id, 'availability', version=version
    )


@test_wrapper
def incorrect_date_format(api_url, api_key, product_id, version=2):
    """Checking incorrect date format"""
    return availability.incorrect_date_format(api_url, api_key, product_id, 'availability', version=version)


@test_wrapper
def end_before_start_error(api_url, api_key, product_id, version=2):
    """Checking incorrect range error"""
    return availability.end_before_start_error(api_url, api_key, product_id, 'availability', version=version)


@test_wrapper
def past_start_date(api_url, api_key, product_id, version=2):
    """Checking past date"""
    return availability.past_start_date(api_url, api_key, product_id, 'availability', version=version)


@test_wrapper
def huge_date_range(api_url, api_key, product_id, version=2):
    """Checking huge date range"""
    return availability.huge_date_range(api_url, api_key, product_id, 'availability', version=version)


@test_wrapper
def not_allowed_method(api_url, api_key, product_id, version=2):
    """Testing methods that are not allowed"""
    return availability.not_allowed_method(api_url, api_key, product_id, 'availability', version=version)



from datetime import datetime, timedelta
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from supplier_api_tester.v2.exceptions import FailedTest
from supplier_api_tester.v2.models import DailyVariants, Timeslot
from supplier_api_tester.v2.models import Product
from supplier_api_tester.v2.utils.availability import get_availability
from supplier_api_tester.v2.utils.catalog import get_catalog


def get_reservation_slot(api_url, api_key, product_id, version=2) -> Optional[DailyVariants]:
    """Getting day of timeslot with at least one variant available."""
    start = datetime.utcnow().date()
    end = start + timedelta(days=30)
    days, raw_response = get_availability(api_url, api_key, product_id, version, start, end)

    days: List[DailyVariants] = [day for day in days if day.variants and day.available_tickets > 0]
    single_variant_item = None
    for day in days:
        variants_max_tickets = [v.available_tickets for v in day.variants if v.available_tickets > 0]
        if len(variants_max_tickets) > 1:
            # got multi variant slot
            return day
        if not single_variant_item:
            single_variant_item = day

    if not single_variant_item:
        raise FailedTest(
            message='There is no availability in the next 30 days to test a reservation.',
            response=raw_response,
        )
    return single_variant_item


def get_payload_for_reservation(
        api_url: str,
        api_key: str,
        product_id: str,
        slot: Union[DailyVariants, Timeslot],
        variant_quantity=1,
        min_quantity=2,
        with_additional_order_data: bool = True,
        with_additional_visitors_data: bool = True,
) -> Dict:
    # get product information to determine the required additional (order, visitors) data
    _, products = get_catalog(api_url, api_key, version=2)
    product = [product for product in products if product.id == product_id][0]
    payload = {
        'datetime': (
            f'{slot.date.isoformat()}T{slot.timeslot}'
            if product.use_timeslots
            else f'{slot.date.isoformat()}T00:00'
        ),
        'customer': {
            'first_name': 'Jon',
            'last_name': 'Snow',
            'email': 'tests@tiqets.com',
            'phone': '+31 85 888 4442',
            'country': 'NL',
        }
    }

    tickets = []

    for variant in slot.variants:
        if variant.available_tickets >= min_quantity:
            ticket = {
                'variant_id': variant.id,
                'quantity': variant_quantity,
            }

            if product.required_visitor_data and with_additional_visitors_data:
                ticket['required_visitor_data'] = get_required_visitors_data_payload(
                    product.required_visitor_data, variant_quantity
                )

            tickets.append(ticket)

    payload['tickets'] = tickets

    if with_additional_order_data:
        required_order_data: Dict = get_required_order_data_payload(product)
        if required_order_data:
            payload['required_order_data'] = required_order_data

    return payload


def get_required_order_data_payload(product: Product) -> Dict:
    if not product.required_order_data:
        return {}

    required_order_data: Dict = {}
    for required_field_name in product.required_order_data:
        # get a (random) value for the field
        required_order_data[required_field_name] = f'test {required_field_name.lower()}'
    return required_order_data


def get_required_visitors_data_payload(required_visitor_data: List[str], variant_quantity: int) -> List[Dict]:
    required_visitors_data: List[Dict] = []
    for i in range(1, variant_quantity + 1):
        data = {}
        for required_field_name in required_visitor_data:
            # get a (random) value for the field
            data[required_field_name] = f'test {required_field_name.lower()} {i}'
        required_visitors_data.append(data)
    return required_visitors_data



import json
import time
from functools import wraps

from .exceptions import FailedTest
from .models import TestResult, Response
from .utils.conversions import ns_to_ms


def test_wrapper(f):
    '''
    This wrapper is doing 3 things:
      - creates the TestResult if the FailedTest exception occurs while running test function
      - adds the test title from the test function docstring
      - measure the test duration in milliseconds
    '''
    @wraps(f)
    def decorated_function(*args, **kwargs):
        start = time.time_ns()
        try:
            test_result = f(*args, **kwargs)
        except FailedTest as e:
            return TestResult(
                title=f.__doc__,
                status=2,
                message=e.message,
                duration=ns_to_ms(time.time_ns() - start),
                response=Response(
                    url=e.response.url,
                    status_code=e.response.status_code,
                    headers=e.response.headers,
                    payload=_format_json(e.response.request.body),
                    body=_format_json(e.response.text),
                ) if e.response is not None else None,
            )
        test_result.title = f.__doc__
        test_result.duration = ns_to_ms(time.time_ns() - start)
        return test_result
    return decorated_function


def _format_json(raw_data):
    if not raw_data:
        return None
    try:
        return json.dumps(json.loads(raw_data), indent=2)
    except:
        return raw_data



from textwrap import wrap
from typing import List
from supplier_api_tester.v2.models import Product

from terminaltables import AsciiTable


class Colors:
    OK = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'


def error_printer(error: str, no_colors: bool = False) -> None:
    rows = [['Error']]
    if no_colors:
        data_row = [
            error,
        ]
    else:
        data_row = [
            f'{Colors.FAIL}{error}{Colors.ENDC}',
        ]
    rows.extend([data_row])
    table = AsciiTable(rows)
    print(table.table)


def results_printer(results, no_colors=False):
    any_message = any(bool(r.message) for r in results)
    headers = ['#', 'Time', 'Test name']
    detailed_report = False
    if no_colors:
        headers.append('Result')
    if any_message:
        headers.append('Description')
    table_data = [headers]
    text_colors = {
        0: Colors.OK,
        1: Colors.WARNING,
        2: Colors.FAIL,
    }
    text_statuses = {
        0: 'OK',
        1: 'WARNINIG',
        2: 'FAIL',
    }
    for i, result in enumerate(results, 1):
        if no_colors:
            data_row = [
                i,
                f'{result.duration}ms',
                result.title,
                text_statuses[result.status],
            ]
        else:
            text_color = text_colors[result.status]
            data_row = [
                i,
                f'{text_color}{result.duration}ms{Colors.ENDC}',
                f'{text_color}{result.title}{Colors.ENDC}',
            ]
        if any_message:
            data_row.append(result.message or '')
        table_data.append(data_row)
        if not detailed_report and result.response:
            detailed_report = True
    table = AsciiTable(table_data)
    if not table.ok:
        error_printer('The terminal is not wide enough to display the results', no_colors)
        return

    if any_message:
        columns_count = len(headers) - 1
        message_size = table.column_max_width(columns_count)
        for column in table.table_data:
            column[columns_count] = '\n'.join(wrap(column[columns_count], message_size))
    print(table.table)

    if detailed_report:
        print('\nREPORTS FROM FAILED TESTS\n')
        for i, result in enumerate(results, 1):
            if result.response:
                if no_colors:
                    print(f'--- #{i} {result.title} ---')
                    print('\nURL')
                    print(result.response.url)
                    print('\nSTATUS_CODE')
                    print(result.response.status_code)
                    print('\nHEADERS')
                    for header, value in result.response.headers.items():
                        print(f'{header}: {value}')
                    print('\nPAYLOAD')
                    print(result.response.payload)
                    print('\nRESPONSE')
                    print(result.response.body)
                else:
                    print(f'{Colors.FAIL}--- #{i} {result.title} ---{Colors.ENDC}')
                    print(f'\n{Colors.WARNING}URL{Colors.ENDC}')
                    print(result.response.url)
                    print(f'\n{Colors.WARNING}STATUS_CODE{Colors.ENDC}')
                    print(result.response.status_code)
                    print(f'\n{Colors.WARNING}HEADERS{Colors.ENDC}')
                    for header, value in result.response.headers.items():
                        print(f'{header}: {value}')
                    print(f'\n{Colors.WARNING}PAYLOAD{Colors.ENDC}')
                    print(result.response.payload)
                    print(f'\n{Colors.WARNING}RESPONSE{Colors.ENDC}')
                    print(result.response.body)
                print()


def products_printer(products: List[Product]) -> None:
    rows = [
        [
            'ID',
            'Name',
            'Timeslots',
            'Refundable',
            'Cutoff time',
            'Required Additional Order Data',
            'Required Additional Visitors Data',
        ]
    ]
    rows.extend([
        [
            p.id,
            p.name,
            p.use_timeslots,
            p.is_refundable,
            p.cutoff_time,
            ','.join(p.required_order_data) if p.required_order_data else '',
            ','.join(p.required_visitor_data) if p.required_visitor_data else '',
        ] for p in products]
    )
    table = AsciiTable(rows)
    print(table.table)



from supplier_api_tester.v2.exceptions import FailedTest
from supplier_api_tester.v2.printer import products_printer
from supplier_api_tester.v2.printer import results_printer
from supplier_api_tester.v2.tester import SupplierApiTester
from supplier_api_tester.v2.utils.catalog import get_catalog


def print_title(title):
    print()
    print('-' * len(title))
    print(title)
    print('-' * len(title))
    print()


def supplier_tester(url, api_key, product_id, availability, reservation, booking, catalog, no_colors):
    """Test your Supplier API implementation"""

    if not any((availability, reservation, booking, catalog)):
        availability = True
        reservation = True
        booking = True
        catalog = True

    if availability:
        print_title('AVAILABILITY TESTS')
        runner = SupplierApiTester(
            host=url,
            api_key=api_key,
            product_id=product_id,
            test_target='availability',
        )
        results = runner.run()
        results_printer(results, no_colors)

    if reservation:
        print_title('RESERVATION TESTS')
        runner = SupplierApiTester(
            host=url,
            api_key=api_key,
            product_id=product_id,
            test_target='reservation',
        )
        results = runner.run()
        results_printer(results, no_colors)

    if booking:
        print_title('BOOKING TESTS')
        runner = SupplierApiTester(
            host=url,
            api_key=api_key,
            product_id=product_id,
            test_target='booking',
        )
        results = runner.run()
        results_printer(results, no_colors)

    if catalog:
        print_title('PRODUCT CATALOG')
        runner = SupplierApiTester(
            host=url,
            api_key=api_key,
            product_id=product_id,
            test_target='catalog',
        )
        results = runner.run()
        results_printer(results, no_colors)


def supplier_products(url: str, api_key: str):
    """Shows the product catalog"""

    products = None
    try:
        print(f'Running CLI tests for API v2.x')
        _, products = get_catalog(url, api_key)
    except FailedTest as e:
        print(f'Unable to get the products: {e.message}')
        exit(1)

    products_printer(products)



from typing import List

from supplier_api_tester.v2.models import TestResult
from supplier_api_tester.v2.tests import (
    AVAILABILITY_TEST,
    BOOKING_TEST,
    PRODUCT_CATALOG,
    RESERVATION_TEST,
)


class SupplierApiTester(object):
    def __init__(
        self,
        host: str,
        api_key: str,
        product_id: str,
        test_target: str,
    ):
        self.host = host
        self.api_key = api_key
        self.product_id = product_id
        self.test_target = test_target

    def run(self) -> List[TestResult]:
        test_targets = {
            'availability': _run_tests(AVAILABILITY_TEST, self.host, self.api_key, self.product_id, 2),
            'reservation': _run_tests(RESERVATION_TEST, self.host, self.api_key, self.product_id, 2),
            'booking': _run_tests(BOOKING_TEST, self.host, self.api_key, self.product_id, 2),
            'catalog': _run_tests(PRODUCT_CATALOG, self.host, self.api_key, 2),
        }
        return list(test_targets[self.test_target])


def _run_tests(target, *args):
    return (test(*args) for test in target)



from requests.models import Response

from supplier_api_tester.v1.exceptions import FailedTest
from supplier_api_tester.v1.models import TestResult


def check_api_error(raw_response: Response, api_error, expected_error) -> TestResult:
    '''Chcking if the API error matches the expected error'''
    if api_error.error_code != expected_error.error_code:
        raise FailedTest(
            message=f'Incorrect error_code ({api_error.error_code}). Expected value: {expected_error.error_code}',
            response=raw_response,
        )
    if api_error.error != expected_error.error:
        raise FailedTest(
            f'Incorrect error text ({api_error.error}). Expected text: {expected_error.error}',
            response=raw_response,
        )
    if not api_error.message.startswith(expected_error.message):
        return TestResult(
            status=1,
            message=f'Incorrect message text "{api_error.message}". Expected text should start with: "{expected_error.message}"',
        )
    return TestResult()



from datetime import date, datetime, timedelta

from typing import List, Tuple

import requests
from requests.models import Response

from supplier_api_tester.v2.client import client
from supplier_api_tester.v2.exceptions import FailedTest
from supplier_api_tester.v2.models import ApiError, DailyVariants, TestResult
from supplier_api_tester.v2.utils.adapters import (
    get_api_error,
    parse_availability_variants,
)
from supplier_api_tester.v2.utils.errors import check_api_error
from supplier_api_tester.v2.utils.date import get_tomorrow


def get_availability(
    api_url: str,
    api_key: str,
    product_id: str,
    version: int,
    start_date: date,
    end_date: date,
) -> Tuple[List[DailyVariants], Response]:
    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/availability', api_key, {
        'start': start_date.isoformat(),
        'end': end_date.isoformat(),
    })
    days: List[DailyVariants] = parse_availability_variants(raw_response, response)
    return days, raw_response


def past_start_date(api_url, api_key, product_id, endpoint, version=2):
    """Checking availability with start date from the past"""
    today = datetime.utcnow().date()
    start = today - timedelta(days=1)
    end = today

    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': start.isoformat(),
        'end': end.isoformat(),
    })
    if not response:
        return TestResult()

    if start in response:
        raise FailedTest(
            message=f'Returned availability for date in the past: {start}',
            response=raw_response,
        )

    return TestResult()


def huge_date_range(api_url, api_key, product_id, endpoint, version=2):
    """Checking availability with huge date range"""
    today = datetime.utcnow().date()
    start = today
    end = today + timedelta(days=365 * 10)

    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': start.isoformat(),
        'end': end.isoformat(),
    })
    return TestResult()


# TODO this no longer needed in v2 because the API error 2009 was removed. Check that the response is not 2009 error
def empty_availability(api_url, api_key, product_id, endpoint, version=2):
    """Checking availability that is supposed to be empty"""
    today = datetime.utcnow().date()
    start = today + timedelta(days=300)
    end = start + timedelta(days=1)

    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': start.isoformat(),
        'end': end.isoformat(),
    })
    if not response:
        return TestResult()
    return TestResult(
        status=1,
        message=(
            "Skipping that test because response is not empty."
        )
    )


def test_missing_api_key(api_url, api_key, product_id, endpoint, version=2):
    """Request without API-Key"""
    tomorrow = get_tomorrow()
    raw_response, _ = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': tomorrow.isoformat(),
    }, headers={})
    if raw_response.status_code != 403:
        raise FailedTest(
            message=f'Incorrect status code "{raw_response.status_code}" when calling the API wihout the API-Key. Expected status code: "403".',
            response=raw_response,
        )

    if raw_response.text != 'Forbidden - Missing or incorrect API key':
        return TestResult(
            status=1,
            message=(
                f'Incorrect text message "{raw_response.text}". '
                'Expected message: "Forbidden - Missing or incorrect API key".'
            )
        )

    return TestResult()


def test_incorrect_api_key(api_url, api_key, product_id, endpoint, version=2):
    """Request with incorrect API-Key"""
    tomorrow = get_tomorrow()
    raw_response, _ = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': tomorrow.isoformat(),
    }, headers={'API-Key': 'NON-EXISTING-API-KEY'})

    if raw_response.status_code != 403:
        raise FailedTest(
            message=f'Incorrect status code "{raw_response.status_code}" when calling the API wihout the API-Key. Expected status code: "403".',
            response=raw_response,
        )

    if raw_response.text != 'Forbidden - Missing or incorrect API key':
        return TestResult(
            status=1,
            message=(
                f'Incorrect text message "{raw_response.text}". '
                'Expected message: "Forbidden - Missing or incorrect API key".'
            )
        )

    return TestResult()


def test_missing_argument_error(api_url, api_key, product_id, endpoint, version=2):
    """Testing missing argument errors"""
    tomorrow = get_tomorrow()
    warnings: List[str] = []

    # end is missing
    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
    })
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message='Required argument "end" was not found',
    )
    result = check_api_error(raw_response, api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    # start is missing
    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message='Required argument "start" was not found',
    )
    result = check_api_error(raw_response, api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    # start and end are missing
    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key)
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message='Required argument "start" was not found',
    )
    result = check_api_error(raw_response, api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    if warnings:
        return TestResult(
            status=1,
            message='\n '.join(warnings)
        )
    return TestResult()


def test_error_for_non_existing_product(api_url, api_key, product_id, endpoint, version=2):
    """Testing availability for non existing product"""
    tomorrow = get_tomorrow()
    raw_response, response = client(f'{api_url}/v{version}/products/NON-EXISTING-PRODUCT-ID/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1001,
        error='Missing product',
        message='Product with ID NON-EXISTING-PRODUCT-ID doesn\'t exist',
    )
    return check_api_error(raw_response, api_error, expected_error)


def incorrect_date_format(api_url, api_key, product_id, endpoint, version=2):
    """Checking incorrect date format"""
    tomorrow = get_tomorrow()
    bad_date_format = tomorrow.strftime('%d-%m-%Y')

    # start date in a bad format
    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': bad_date_format,
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=2000,
        error='Malformed datetime',
        message=f'Incorrect date format {bad_date_format}, please use the YYYY-MM-DD format',
    )
    check_api_error(raw_response, api_error, expected_error)

    # end date in a bad format
    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': bad_date_format,
    })
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=2000,
        error='Malformed datetime',
        message=f'Incorrect date format {bad_date_format}, please use the YYYY-MM-DD format',
    )
    check_api_error(raw_response, api_error, expected_error)

    return TestResult()


def end_before_start_error(api_url, api_key, product_id, endpoint, version=2):
    """Checking incorrect range error"""
    tomorrow = get_tomorrow()
    next_week = tomorrow + timedelta(days=7)
    raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': next_week.isoformat(),
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=2001,
        error='Incorrect date range',
        message='The end date cannot be earlier than start date',
    )
    return check_api_error(raw_response, api_error, expected_error)


def not_allowed_method(api_url, api_key, product_id, endpoint, version=2):
    """Testing methods that are not allowed"""
    tomorrow = get_tomorrow()
    for method in (requests.post, requests.put, requests.patch, requests.delete):
        raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
            'start': tomorrow.isoformat(),
            'end': tomorrow.isoformat(),
        }, method=method)
        status_code = getattr(raw_response, 'status_code', 200)
        if status_code != 405:
            raise FailedTest(
                message=f'Incorrect status code "{status_code}" when calling the API via method {method.__name__.upper()}. Expected status code: "405".',
                response=raw_response,
            )
    return TestResult()


# def test_error_for_non_timeslot_product(api_url, api_key, product_id, endpoint, version=2):
#     '''Testing timeslot availability for non timeslot product'''
#     tomorrow = get_tomorrow()
#     raw_response, response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
#         'start': tomorrow.isoformat(),
#         'end': tomorrow.isoformat(),
#     })
#     api_error = get_api_error(raw_response, response)
#     expected_error = ApiError(
#         error_code=1002,
#         error='Timeslot product expected',
#         message=f'Requested timeslot availability for non timeslot product ID {product_id}',
#     )
#     return check_api_error(raw_response, api_error, expected_error)



import click

from supplier_api_tester.v2.cli import supplier_tester as supplier_tester_v2
from supplier_api_tester.v2.cli import supplier_products as supplier_products_v2

from supplier_api_tester.v1.cli import supplier_tester as supplier_tester_v1
from supplier_api_tester.v1.cli import supplier_products as supplier_products_v1


@click.command()
@click.option('-u', '--url', required=True, prompt='Server URL', type=str)
@click.option('-k', '--api-key', required=True, prompt='API Key', type=str)
@click.option('-p', '--product-id', type=str, help='Product ID to call tests on. Required with -a and -t flags')
@click.option('-t', '--timeslots', is_flag=True, default=False, help='Use timeslots. Not supported in v2.x')
@click.option('-a', '--availability', is_flag=True, default=False, help='Run availability tests')
@click.option('-r', '--reservation', is_flag=True, default=False, help='Run reservation tests')
@click.option('-b', '--booking', is_flag=True, default=False, help='Run booking tests')
@click.option('-c', '--catalog', is_flag=True, default=False, help='Run product catalog tests')
@click.option('-nc', '--no-colors', is_flag=True, default=False, help='Not using colors on output')
@click.option('-v', '--version', default=2, help='Choosing the API version', prompt='API version', type=int)
def supplier_tester(
        url, api_key, product_id, timeslots, availability, reservation, booking, catalog, no_colors, version: int
):
    """Test your Supplier API implementation"""

    if not any((availability, reservation, booking, catalog)):
        availability = True
        reservation = True
        booking = True
        catalog = True

    if any((availability, timeslots, reservation, booking)) and not product_id:
        product_id = click.prompt('Product ID')

    if version == 1:
        supplier_tester_v1(url, api_key, product_id, timeslots, availability, reservation, booking, catalog, no_colors)
    else:
        supplier_tester_v2(url, api_key, product_id, availability, reservation, booking, catalog, no_colors)


@click.command()
@click.option('-u', '--url', required=True, prompt='Server URL', type=str)
@click.option('-k', '--api-key', required=True, prompt='API Key', type=str)
@click.option('-v', '--version', default=2, help='Choosing the API version', prompt='API version', type=int)
def supplier_products(url: str, api_key: str, version: int):
    """Shows the product catalog"""

    if version == 1:
        supplier_products_v1(url, api_key)
    else:
        supplier_products_v2(url, api_key)



# If multiple checks can be done using the same response
# then they should be done under a single test case.

from typing import List, Optional, Tuple
import requests
from requests.models import Response

from ..client import client
from ..decorators import test_wrapper
from ..exceptions import FailedTest
from ..models import Product, TestResult
from ..utils.adapters import get_products


def get_catalog(
    api_url: str,
    api_key: str,
    version: int,
    use_timeslots: Optional[bool] = None,
) -> Tuple[Response, List[Product]]:
    url = f'{api_url}/v{version}/products'
    params = {}
    if use_timeslots is not None:
        params['use_timeslots'] = use_timeslots
    raw_response, response = client(url, api_key, method=requests.get, params=params)
    products = get_products(raw_response, response)
    return raw_response, products


@test_wrapper
def test_get_products(api_url, api_key, version=1):
    '''Get product catalog'''
    get_catalog(api_url, api_key, version)
    return TestResult()


@test_wrapper
def test_get_timeslots_products(api_url, api_key, version=1):
    '''Get product catalog with use_timeslots=True query filter'''
    raw_response, products = get_catalog(api_url, api_key, version, use_timeslots=True)
    _validate_timeslots(raw_response, products, use_timeslots=True)
    return TestResult()


@test_wrapper
def test_get_no_timeslots_products(api_url, api_key, version=1):
    '''Get product catalog with use_timeslots=False query filter'''
    raw_response, products = get_catalog(api_url, api_key, version, use_timeslots=False)
    _validate_timeslots(raw_response, products, use_timeslots=False)
    return TestResult()


def _validate_timeslots(raw_response: Response, products: List[Product], use_timeslots: bool):
    for product in products:
        if product.use_timeslots != use_timeslots:
            raise FailedTest(
                message=f'Product {product.id} with non matching use_timeslots returned',
                response=raw_response,
            )



from typing import List, Tuple
import requests
from requests.models import Response

from supplier_api_tester.v2.client import client
from supplier_api_tester.v2.models import Product
from ..utils.adapters import get_products


def get_catalog(
        api_url: str,
        api_key: str,
        version: int = 2,
) -> Tuple[Response, List[Product]]:
    url = f'{api_url}/v{version}/products'
    raw_response, response = client(url, api_key, method=requests.get, params={})
    products = get_products(raw_response, response)
    return raw_response, products



from requests.models import Response


class FailedTest(Exception):
    message: str = None
    response: Response = None

    def __init__(self, message: str, response: Response = None):
        self.message = message
        self.response = response



from typing import List

from .models import TestResult
from .tests import (
    AVAILABILITY_TEST,
    AVAILABILITY_VARIANTS_TEST,
    AVAILABILITY_TIMESLOTS_TEST,
    BOOKING_TEST,
    PRODUCT_CATALOG,
    RESERVATION_TEST,
)


class SupplierApiTesterV1(object):
    def __init__(
        self,
        host: str,
        api_key: str,
        product_id: str,
        test_target: str,
        timeslots: bool,
    ):
        self.host = host
        self.api_key = api_key
        self.product_id = product_id
        self.test_target = test_target
        self.timeslots = timeslots

    def run(self) -> List[TestResult]:
        test_targets = {
            'availability': _run_tests(AVAILABILITY_TEST, self.host, self.api_key, self.product_id),
            'reservation': _run_tests(RESERVATION_TEST, self.host, self.api_key, self.product_id, self.timeslots),
            'booking': _run_tests(BOOKING_TEST, self.host, self.api_key, self.product_id, self.timeslots),
            'catalog': _run_tests(PRODUCT_CATALOG, self.host, self.api_key),
            'timeslots': _run_tests(AVAILABILITY_TIMESLOTS_TEST, self.host, self.api_key, self.product_id),
            'variants': _run_tests(AVAILABILITY_VARIANTS_TEST, self.host, self.api_key, self.product_id),
        }
        results = list(test_targets[self.test_target])
        if self.test_target == 'availability':
            if self.timeslots:
                return results + list(test_targets['timeslots'])
            else:
                return results + list(test_targets['variants'])
        return results


def _run_tests(target, *args):
    return (test(*args) for test in target)



# If multiple checks can be done using the same response
# then they should be done under a single test case.

from datetime import datetime, timedelta, timezone
from typing import List

import requests

from supplier_api_tester.v2.client import client
from supplier_api_tester.v2.decorators import test_wrapper
from supplier_api_tester.v2.exceptions import FailedTest
from supplier_api_tester.v2.models import TestResult, ApiError
from ..utils.adapters import get_api_error
from ..utils.date import get_tomorrow
from ..utils.adapters import get_reservation
from ..utils.errors import check_api_error
from ..utils.reservation import get_payload_for_reservation
from ..utils.reservation import get_reservation_slot


@test_wrapper
def test_missing_api_key(api_url, api_key, product_id, version=2):
    """Request without API-Key"""
    raw_response, _ = client(f'{api_url}/v{version}/products/{product_id}/reservation', api_key, method=requests.post, json_payload={}, headers={})

    if raw_response.status_code != 403:
        raise FailedTest(
            message=f'Incorrect status code "{raw_response.status_code}" when calling the API wihout the API-Key. Expected status code: "403".',
            response=raw_response,
        )

    if raw_response.text != 'Forbidden - Missing or incorrect API key':
        return TestResult(
            status=1,
            message=(
                f'Incorrect text message "{raw_response.text}". '
                'Expected message: "Forbidden - Missing or incorrect API key".'
            )
        )

    return TestResult()


@test_wrapper
def test_incorrect_api_key(api_url, api_key, product_id, version=2):
    """Request with incorrect API-Key"""
    raw_response, _ = client(f'{api_url}/v{version}/products/{product_id}/reservation', api_key, method=requests.post, json_payload={}, headers={
        'API-Key': 'NON-EXISTING-API-KEY',
    })

    if raw_response.status_code != 403:
        raise FailedTest(
            message=f'Incorrect status code "{raw_response.status_code}" when calling the API wihout the API-Key. Expected status code: "403".',
            response=raw_response,
        )

    if raw_response.text != 'Forbidden - Missing or incorrect API key':
        return TestResult(
            status=1,
            message=(
                f'Incorrect text message "{raw_response.text}". '
                'Expected message: "Forbidden - Missing or incorrect API key".'
            )
        )

    return TestResult()


@test_wrapper
def test_missing_argument_error(api_url, api_key, product_id, version=2):
    """Testing missing argument errors"""
    tomorrow = get_tomorrow()
    warnings: List[str] = []
    slot = get_reservation_slot(api_url, api_key, product_id)

    # payload is empty
    json_payload = {}
    raw_response, response = client(
        f'{api_url}/v{version}/products/{product_id}/reservation',
        api_key,
        method=requests.post,
        json_payload=json_payload,
    )
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message='Required argument "datetime" was not found',
    )
    result = check_api_error(raw_response, api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    # no tickets
    json_payload['datetime'] = f'{tomorrow.isoformat()}T00:00'
    raw_response, response = client(
        f'{api_url}/v{version}/products/{product_id}/reservation',
        api_key,
        method=requests.post,
        json_payload=json_payload,
    )
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message='Required argument "tickets" was not found',
    )
    result = check_api_error(raw_response, api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    # no customer
    json_payload['tickets'] = [{
        'variant_id': slot.variants[0].id,
        'quantity': 1,
    }]
    raw_response, response = client(
        f'{api_url}/v{version}/products/{product_id}/reservation',
        api_key,
        method=requests.post,
        json_payload=json_payload,
    )
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message='Required argument "customer" was not found',
    )
    result = check_api_error(raw_response, api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    if warnings:
        return TestResult(
            status=1,
            message='\n '.join(warnings)
        )
    return TestResult()


@test_wrapper
def test_error_for_non_existing_product(api_url, api_key, product_id, version=2):
    """Testing reservation for non-existing product"""
    url = f'{api_url}/v{version}/products/NON-EXISTING-PRODUCT-ID/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id)
    json_payload = get_payload_for_reservation(api_url, api_key, product_id, slot)
    raw_response, response = client(url, api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1001,
        error='Missing product',
        message='Product with ID NON-EXISTING-PRODUCT-ID does not exist',
    )
    check_api_error(raw_response, api_error, expected_error)
    return TestResult()


@test_wrapper
def test_incorrect_date_format(api_url, api_key, product_id, version=2):
    """Testing reservation with incorrect date format"""
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    bad_date_format = '05/05/2020'
    slot = get_reservation_slot(api_url, api_key, product_id)
    json_payload = get_payload_for_reservation(api_url, api_key, product_id, slot)
    json_payload['datetime'] = bad_date_format
    raw_response, response = client(url, api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=2000,
        error='Malformed datetime',
        message=f'Incorrect date format {bad_date_format}, please use the YYYY-MM-DDTHH:MM format',
    )
    check_api_error(raw_response, api_error, expected_error)
    return TestResult()


@test_wrapper
def test_past_date(api_url, api_key, product_id, version=2):
    """Testing reservation with past date"""
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id)
    json_payload = get_payload_for_reservation(api_url, api_key, product_id, slot)
    yesterday = datetime.utcnow().date() - timedelta(days=1)
    json_payload['datetime'] = f'{yesterday.isoformat()}T00:00'
    raw_response, response = client(url, api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=2009,
        error='Incorrect date',
        message=f'Cannot use the past date',
    )
    check_api_error(raw_response, api_error, expected_error)
    return TestResult()


@test_wrapper
def test_not_allowed_method(api_url, api_key, product_id, version=2):
    """Testing methods that are not allowed"""
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id)
    json_payload = get_payload_for_reservation(api_url, api_key, product_id, slot)
    for method in (requests.get, requests.put, requests.patch, requests.delete):
        raw_response, _ = client(url, api_key, method=method, json_payload=json_payload)
        status_code = getattr(raw_response, 'status_code', 200)
        if status_code != 405:
            raise FailedTest(
                message=f'Incorrect status code "{status_code}" when calling the API via method {method.__name__.upper()}. Expected status code: "405".',
                response=raw_response,
            )
    return TestResult()


@test_wrapper
def test_reservation(api_url, api_key, product_id, version=2):
    """Reserving tickets for at least 1 variant"""
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id)
    json_payload = get_payload_for_reservation(api_url, api_key, product_id, slot)
    raw_response, response = client(url, api_key, method=requests.post, json_payload=json_payload)
    reservation = get_reservation(raw_response, response)
    if not reservation.reservation_id:
        raise FailedTest(
            message='No reservation ID found',
            response=raw_response,
        )
    if not reservation.expires_at.tzname():
        raise FailedTest(
            message='Expiration time should have the timezone.',
            response=raw_response,
        )
    if datetime.now(timezone.utc) + timedelta(minutes=14) > reservation.expires_at:
        raise FailedTest(
            message='Reservation should be held at least 15 minutes.',
            response=raw_response,
        )
    return TestResult()



from datetime import datetime, timedelta
from typing import Union

from supplier_api_tester.v1.exceptions import FailedTest
from supplier_api_tester.v1.models import DailyVariants, Timeslot
from supplier_api_tester.v1.utils.availability import (
    get_availability_timeslots,
    get_availability_variants,
)


def get_reservation_slot(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Getting day of timeslot with at least one variant available.'''
    start = datetime.utcnow().date()
    end = start + timedelta(days=30)
    if timeslots:
        days, raw_response = get_availability_timeslots(api_url, api_key, product_id, version, start, end)
    else:
        days, raw_response = get_availability_variants(api_url, api_key, product_id, version, start, end)

    days = [day for day in days if day.variants and day.max_tickets > 0]
    single_variant_item = None
    for day in days:
        variants_max_tickets = [v.max_tickets for v in day.variants if v.max_tickets > 0]
        if len(variants_max_tickets) > 1:
            # got multi variant slot
            return day
        if not single_variant_item:
            single_variant_item = day

    if not single_variant_item:
        raise FailedTest(
            message='There is no availability in the next 30 days to test a reservation.',
            response=raw_response,
        )
    return single_variant_item


def get_payload_from_slot(slot: Union[DailyVariants, Timeslot], variant_quantity=1, min_quantity=1):
    return {
        'date': slot.date.isoformat(),
        'tickets': [{
            'variant_id': variant.id,
            'quantity': variant_quantity,
        } for variant in slot.variants if variant.max_tickets >= min_quantity],
        'customer': {
            'first_name': 'Jon',
            'last_name': 'Snow',
            'email': 'tests@tiqets.com',
            'phone': '+31 85 888 4442',
            'country': 'NL',
        }
    }



# If multiple checks can be done using the same response
# then they should be done under a single test case.

from datetime import datetime
import time

import requests

from supplier_api_tester.v2.client import client
from supplier_api_tester.v2.decorators import test_wrapper
from supplier_api_tester.v2.exceptions import FailedTest
from supplier_api_tester.v2.models import TestResult, ApiError
from ..utils.adapters import get_reservation, get_booking, get_api_error
from ..utils.catalog import get_catalog
from ..utils.reservation import get_payload_for_reservation
from ..utils.reservation import get_reservation_slot
from ..utils.errors import check_api_error


def reference_id() -> str:
    return str(time.time_ns())


@test_wrapper
def test_missing_reservation_id(api_url, api_key, product_id, version=2):
    """Booking without the reservation ID"""
    url = f'{api_url}/v{version}/booking'
    raw_response, response = client(url, api_key, method=requests.post, json_payload={})
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message='Required argument "reservation_id" was not found',
    )
    check_api_error(raw_response, api_error, expected_error)
    return TestResult()


@test_wrapper
def test_missing_api_key(api_url, api_key, product_id, version=2):
    """Booking without the API key"""
    url = f'{api_url}/v{version}/booking'
    raw_response, _ = client(url, api_key, method=requests.post, json_payload={}, headers={})

    if raw_response.status_code != 403:
        raise FailedTest(
            message=f'Incorrect status code "{raw_response.status_code}" when calling the API wihout the API-Key. Expected status code: "403".',
            response=raw_response,
        )

    if raw_response.text != 'Forbidden - Missing or incorrect API key':
        return TestResult(
            status=1,
            message=(
                f'Incorrect text message "{raw_response.text}". '
                'Expected message: "Forbidden - Missing or incorrect API key".'
            )
        )
    return TestResult()


@test_wrapper
def test_incorrect_api_key(api_url, api_key, product_id, version=2):
    """Booking with incorrect API-Key"""
    url = f'{api_url}/v{version}/booking'
    raw_response, _ = client(
        url, api_key, method=requests.post, json_payload={}, headers={'API-Key': 'NON-EXISTING-API-KEY'}
    )

    if raw_response.status_code != 403:
        raise FailedTest(
            message=f'Incorrect status code "{raw_response.status_code}" when calling the API wihout the API-Key. Expected status code: "403".',
            response=raw_response,
        )

    if raw_response.text != 'Forbidden - Missing or incorrect API key':
        return TestResult(
            status=1,
            message=(
                f'Incorrect text message "{raw_response.text}". '
                'Expected message: "Forbidden - Missing or incorrect API key".'
            )
        )

    return TestResult()


@test_wrapper
def test_not_allowed_method(api_url, api_key, product_id, version=2):
    """Testing methods that are not allowed"""
    url = f'{api_url}/v{version}/booking'
    for method in (requests.get, requests.put, requests.patch, requests.delete):
        raw_response, _ = client(url, api_key, method=method, json_payload={})
        status_code = getattr(raw_response, 'status_code', 200)
        if status_code != 405:
            raise FailedTest(
                message=f'Incorrect status code "{status_code}" when calling the API via method {method.__name__.upper()}. Expected status code: "405".',
                response=raw_response,
            )
    return TestResult()


@test_wrapper
def test_booking_incorrect_reservation_id(api_url, api_key, product_id, version=2):
    """Booking with incorrect reservation ID."""
    url = f'{api_url}/v{version}/booking'
    raw_response, response = client(url, api_key, method=requests.post, json_payload={
        'reservation_id': 'non-existing-ID',
        'order_reference': reference_id(),
    })
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=3002,
        error='Incorrect reservation ID',
        message="Given reservation ID is incorrect",
    )
    check_api_error(raw_response, api_error, expected_error)
    return TestResult()


@test_wrapper
def test_booking(api_url, api_key, product_id, version=2):
    """Booking tickets for at least 1 variant"""
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id)
    variant_quantity_map = {
         variant.id: 2 for variant in slot.variants if variant.available_tickets > 2
    }

    json_payload = get_payload_for_reservation(api_url, api_key, product_id, slot, variant_quantity=2, min_quantity=3)
    raw_response, response = client(url, api_key, method=requests.post, json_payload=json_payload)
    reservation = get_reservation(raw_response, response)

    url = f'{api_url}/v{version}/booking'
    raw_response, response = client(url, api_key, method=requests.post, json_payload={
        'reservation_id': reservation.reservation_id,
        'order_reference': reference_id(),
    })
    booking = get_booking(raw_response, response)
    if booking.barcode_scope == 'ticket':
        for variant_id, tickets_quantity in variant_quantity_map.items():
            tickets_for_variant = booking.tickets.get(variant_id)
            if tickets_for_variant is None:
                raise FailedTest(
                    message=f'No tickets for variant {variant_id}',
                    response=raw_response,
                )
            if len(tickets_for_variant) != tickets_quantity:
                raise FailedTest(
                    message=
                    f'Expected {tickets_quantity} codes for variant {variant_id} but got {len(tickets_for_variant)}',
                    response=raw_response,
                )

    return TestResult()


@test_wrapper
def test_cancellation(api_url, api_key, product_id, version=2):
    """Perform booking that will be cancelled"""
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id)

    # make a reservation
    json_payload = get_payload_for_reservation(api_url, api_key, product_id, slot, variant_quantity=2, min_quantity=3)
    raw_response, response = client(url, api_key, method=requests.post, json_payload=json_payload)
    reservation = get_reservation(raw_response, response)

    # make a booking
    url = f'{api_url}/v{version}/booking'
    raw_response, response = client(url, api_key, method=requests.post, json_payload={
        'reservation_id': reservation.reservation_id,
        'order_reference': reference_id(),
    })
    booking = get_booking(raw_response, response)
    booking_id = booking.booking_id

    # get product's information to check if product supports cancellations
    _, products = get_catalog(api_url, api_key, version)
    product = [product for product in products if product.id == product_id][0]

    # cancel existing booking
    url = f'{api_url}/v{version}/booking/{booking_id}'
    raw_response, response = client(url, api_key, method=requests.delete)

    # throw error if product does not support cancellation
    if not product.is_refundable:
        api_error = get_api_error(raw_response, response)
        expected_error = ApiError(
            error_code=3004,
            error='Cancellation not possible',
            message='The booking cannot be cancelled, the product does not allow cancellations',
        )
        check_api_error(raw_response, api_error, expected_error)
        return TestResult(
            status=1,
            message=(
                "Skipping that test because the product does not support cancellations"
            )
        )

    # cancel booking in refundable product that cannot be cancelled due to cut_off time/being past date
    if product.use_timeslots:
        booking_for_time = datetime.fromisoformat(f'{slot.date.isoformat()} {slot.timeslot}')
    else:
        booking_for_time = datetime.fromisoformat(slot.date.isoformat())
    cancellation_time = datetime.utcnow()
    if booking_for_time < cancellation_time:
        api_error = get_api_error(raw_response, response)
        expected_error = ApiError(
            error_code=2009,
            error='Incorrect date',
            message='Cannot use the past date',
        )
        check_api_error(raw_response, api_error, expected_error)

    difference = booking_for_time - cancellation_time
    hours_in_advance = round(difference.total_seconds()/3600)
    if product.cutoff_time != 0 and product.cutoff_time > hours_in_advance:
        api_error = get_api_error(raw_response, response)
        expected_error = ApiError(
            error_code=2009,
            error='Incorrect date',
            message=f'The booking can only be cancelled {product.cutoff_time} hours in advance',
        )
        check_api_error(raw_response, api_error, expected_error)

    # cancel booking that was already cancelled:
    url = f'{api_url}/v{version}/booking/{booking_id}'
    raw_response, response = client(url, api_key, method=requests.delete)

    cancel_date_ok = (booking_for_time > cancellation_time)
    cancelled_before_cutoff = (product.cutoff_time == 0 or product.cutoff_time < hours_in_advance)
    cancellation_time_ok = (cancel_date_ok and cancelled_before_cutoff)
    if cancellation_time_ok and product.is_refundable:
        api_error = get_api_error(raw_response, response)
        expected_error = ApiError(
            error_code=3003,
            error='Already cancelled',
            message=f'The booking with ID {booking_id} was already cancelled',
        )
        check_api_error(raw_response, api_error, expected_error)

    # cancel booking with no ID/non-existent ID
    non_existing_booking_id = "I-DO-NOT-EXIST"
    url = f'{api_url}/v{version}/booking/{non_existing_booking_id}'
    raw_response, response = client(url, api_key, method=requests.delete)
    api_error = get_api_error(raw_response, response)
    expected_error = ApiError(
        error_code=1004,
        error='Missing booking',
        message=f"Booking with ID {booking_id} doesn't exist",
    )
    check_api_error(raw_response, api_error, expected_error)
    return TestResult()



import json
import requests
from requests.models import Response

from typing import Any, Callable, Tuple

from supplier_api_tester.v2.exceptions import FailedTest


def client(
    url: str,
    api_key: str,
    params: dict = None,
    headers: dict = None,
    method: Callable = requests.get,
    json_payload=None,
) -> Tuple[Response, Any]:
    params = params if params is not None else {}
    headers = headers if headers is not None else {'API-Key': api_key, 'TIQETS-TEST-ORDER': 'true'}
    headers['Cache-Control'] = 'no-store,no-cache,must-revalidate,max-age=0,post-check=0,pre-check=0'
    headers['Pragma'] = 'no-cache'
    try:
        response = method(url, params=params, headers=headers, json=json_payload)
    except requests.exceptions.ConnectionError:
        raise FailedTest(
            message=f'Connection error occurred while testing endpoint {url}. Check if your server is available.',
            response=None,
        )
    except requests.exceptions.HTTPError:
        raise FailedTest(
            message=f'HTTP error occurred while testing endpoint {url}',
            response=None,
        )

    if response.status_code not in (200, 204, 400, 403, 405, 500):
        raise FailedTest(
            message=f'Unexpected status code {response.status_code} from {url}',
            response=response,
        )

    if response.status_code in (204, 403, 405, 500):
        return response, None

    try:
        return response, response.json()
    except json.JSONDecodeError:
        raise FailedTest(
            message=f'Failed to decode JSON',
            response=response,
        )



import setuptools

from supplier_api_tester.version import __version__

with open('README.md') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name='Supplier API tester',
    version=__version__,
    author='Tiqets Team',
    author_email='connections@tiqets.com',
    description='Console tool for validating the Supplier API implementation',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'supplier_tester=supplier_api_tester.cli:supplier_tester',
            'supplier_products=supplier_api_tester.cli:supplier_products',
        ],
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
    install_requires=requirements,
)



from dataclasses import dataclass
from datetime import date, datetime
from typing import Optional, List, Dict


@dataclass
class Response:
    url: str
    status_code: int
    headers: Dict[str, str]
    payload: Optional[str] = None
    body: Optional[str] = None


@dataclass
class TestResult:
    title: str = ''
    status: int = 0
    message: Optional[str] = None
    duration: int = 0
    response: Optional[Response] = None

    @property
    def status_text(self):
        return {
            0: 'OK',
            1: 'WARNING',
            2: 'ERROR',
        }[self.status]

    @property
    def is_ok(self):
        return self.status == 0

    @property
    def is_warning(self):
        return self.status == 1

    @property
    def is_error(self):
        return self.status == 2


@dataclass
class Product:
    id: str
    name: str
    use_timeslots: bool
    description: Optional[str]
    is_refundable: bool
    cutoff_time: int
    required_order_data: Optional[List[str]] = None
    required_visitor_data: Optional[List[str]] = None


@dataclass
class VariantPrice:
    currency: str
    face_value: str


@dataclass
class Variant:
    id: str
    name: str
    available_tickets: int
    price: VariantPrice


@dataclass
class DailyVariants:
    date: date
    timeslot: str
    available_tickets: int
    variants: List[Variant]


@dataclass
class Timeslot:
    date: date
    start: str
    end: str
    max_tickets: int
    variants: List[Variant]


@dataclass
class ApiError:
    error: str
    error_code: int
    message: str


@dataclass
class Reservation:
    reservation_id: str
    expires_at: datetime


@dataclass
class Booking:
    booking_id: str
    barcode_format: str
    barcode_scope: str
    barcode: Optional[str]
    tickets: Optional[Dict[str, List[str]]]



from . import availability_variants
from . import booking
from . import product_catalog
from . import reservation

AVAILABILITY_TEST = (
    availability_variants.test_next_30_days,
    availability_variants.test_missing_api_key,
    availability_variants.test_incorrect_api_key,
    availability_variants.test_missing_argument_error,
    availability_variants.test_error_for_non_existing_product,
    availability_variants.incorrect_date_format,
    availability_variants.end_before_start_error,
    availability_variants.past_start_date,
    availability_variants.huge_date_range,
    availability_variants.not_allowed_method,
)

RESERVATION_TEST = (
    reservation.test_missing_api_key,
    reservation.test_incorrect_api_key,
    reservation.test_missing_argument_error,
    reservation.test_reservation,
    reservation.test_error_for_non_existing_product,
    reservation.test_incorrect_date_format,
    reservation.test_past_date,
    reservation.test_not_allowed_method,
)

BOOKING_TEST = (
    booking.test_missing_reservation_id,
    booking.test_missing_api_key,
    booking.test_incorrect_api_key,
    booking.test_not_allowed_method,
    booking.test_booking_incorrect_reservation_id,
    booking.test_booking,
    booking.test_cancellation,
)

PRODUCT_CATALOG = (
    product_catalog.test_get_products,
)



# If multiple checks can be done using the same response
# then they should be done under a single test case.

from typing import List
from requests.models import Response

from supplier_api_tester.v2.decorators import test_wrapper
from supplier_api_tester.v2.exceptions import FailedTest
from supplier_api_tester.v2.models import Product, TestResult
from ..utils.catalog import get_catalog


@test_wrapper
def test_get_products(api_url, api_key, version=2):
    """Get product catalog"""
    get_catalog(api_url, api_key, version)
    return TestResult()


def _validate_timeslots(raw_response: Response, products: List[Product], use_timeslots: bool):
    for product in products:
        if product.use_timeslots != use_timeslots:
            raise FailedTest(
                message=f'Product {product.id} with non matching use_timeslots returned',
                response=raw_response,
            )



import base64
import binascii

from datetime import date, datetime
from typing import List

import dacite
from requests.models import Response

from ..exceptions import FailedTest
from ..models import ApiError, Booking, DailyVariants, Product, Reservation, Timeslot


def check_base64(item):
    try:
        return base64.b64encode(base64.b64decode(item)).decode() == item
    except binascii.Error:
        return False


def format_error_message(e: Exception) -> str:
    '''
    Method for enhancing the error messages from Dacite.
    '''
    if isinstance(e, dacite.exceptions.UnexpectedDataError):
        return str(e).replace(
            'to any data class field',
            'to any field from the specification'
        )
    return str(e)


def booking_pdf_validator(booking: Booking, raw_response: Response):
    if booking.barcode_position == 'order':
        if not check_base64(booking.barcode):
            raise FailedTest(
                message = "Error while decoding (base64) PDF voucher for the order",
                response=raw_response
        )
    elif booking.barcode_position == 'ticket':
        for variant, barcodes in booking.tickets.items():
            for barcode in barcodes:
                if not check_base64(barcode):
                    raise FailedTest(
                        message="Error while decoding (base64) PDF voucher for the ticket",
                        response=raw_response
                    )


def parse_availability_variants(raw_response: Response, response) -> List[DailyVariants]:
    '''Getting and testing response from the /variants endpoint'''
    if type(response) is not list:
        raise FailedTest(
            message='The response should be a JSON Array',
            response=raw_response,
        )
    try:
        days = [
            dacite.from_dict(
                data_class=DailyVariants,
                data=day,
                config=dacite.Config(
                    type_hooks={date: date.fromisoformat},
                    strict=True,
                )
            )
            for day in response
        ]
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /variants endpoint: {format_error_message(e)}',
            response=raw_response,
        )
    return days


def parse_availability_timeslots(raw_response: Response, response) -> List[Timeslot]:
    '''Getting and testing response from the /timeslots endpoint'''
    if type(response) is not list:
        raise FailedTest(
            message='The response should be a JSON Array',
            response=raw_response,
        )
    try:
        timeslots = [
            dacite.from_dict(
                data_class=Timeslot,
                data=timeslot,
                config=dacite.Config(
                    type_hooks={date: date.fromisoformat},
                    strict=True,
                )
            )
            for timeslot in response
        ]
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /timeslots endpoint: {format_error_message(e)}',
            response=raw_response,
        )
    return timeslots


def get_products(raw_response: Response, response) -> List[Product]:
    '''Getting and testing response from the /products endpoint'''
    if type(response) is not list:
        raise FailedTest(
            message='The response should be a JSON Array',
            response=raw_response,
        )
    try:
        products = [
            dacite.from_dict(
                data_class=Product,
                data=product,
            )
            for product in response
        ]
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /products endpoint: {format_error_message(e)}',
            response=raw_response,
        )
    return products


def get_reservation(raw_response: Response, response) -> Reservation:
    '''Getting and testing response from the /reservation endpoint'''
    if type(response) is not dict:
        raise FailedTest(
            message='The response should be a JSON Object',
            response=raw_response,
        )
    try:
        return dacite.from_dict(
            data_class=Reservation,
            data=response,
            config=dacite.Config(
                type_hooks={datetime: datetime.fromisoformat},
                strict=True,
            )
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /reservation endpoint: {format_error_message(e)}',
            response=raw_response,
        )


def get_booking(raw_response: Response, response) -> Booking:
    '''Getting and testing response from the /booking endpoint'''
    if type(response) is not dict:
        raise FailedTest(
            message='The response should be a JSON Object',
            response=raw_response,
        )
    try:
        booking = dacite.from_dict(
            data_class=Booking,
            data=response,
            config=dacite.Config(strict=True)
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect JSON format in response from the /booking endpoint: {format_error_message(e)}',
            response=raw_response,
        )

    if booking.barcode_format not in ('QRCODE', 'CODE128', 'CODE39', 'ITF', 'DATAMATRIX', 'EAN13', 'PDF'):

        raise FailedTest(
            message=f'Incorrect barcode format ({booking.barcode_format})',
            response=raw_response,
        )

    if booking.barcode_format == 'PDF':
        booking_pdf_validator(booking, raw_response)

    if booking.barcode_position not in ('order', 'ticket'):
        raise FailedTest(
            message=f'Incorrect value in the barcode_position field: {booking.barcode_position}',
            response=raw_response,
        )
    if booking.barcode_position == 'order' and not booking.barcode:
        raise FailedTest(
            message='Barcode for the whole order is empty',
            response=raw_response,
        )
    if booking.barcode_position == 'ticket' and not booking.tickets:
        raise FailedTest(
            message='Tickets Array is empty',
            response=raw_response,
        )
    return booking


def get_api_error(raw_response: Response, response) -> ApiError:
    """Unpacking 400 error JSON structure"""
    if raw_response.ok:
        raise FailedTest(
            message=f'Expected HTTP 400 error but got {raw_response.status_code}',
            response=raw_response,
        )
    if type(response) is not dict:
        raise FailedTest(
            message='400 error response should be a JSON Object',
            response=raw_response,
        )
    try:
        return dacite.from_dict(
            data_class=ApiError,
            data=response,
            config=dacite.Config(strict=True),
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ) as e:
        raise FailedTest(
            message=f'Incorrect response format for 400 error: {format_error_message(e)}',
            response=raw_response,
        )
